# license
# block
# goes
# here

import bpy
import bpy_extras
from bpy.props import (
    BoolProperty,
    FloatProperty,
    StringProperty,
    IntProperty,
    EnumProperty
    # CollectionProperty
    )
from bpy_extras.io_utils import (
    ExportHelper,
    # orientation_helper,
    # path_reference_mode,
    axis_conversion,
    )
from bpy_extras.object_utils import (
    AddObjectHelper
    )

import inspect
from os import path

bl_info = {
    "name": "Blender2MaxWhere",
    "author": "MISTEMS Ltd",
    "version": (1, 2, 7),
    "blender": (2, 90, 0),
    "location": "File > Export...",
    "description": "Export MaxWhere compatible scenes or files",
    "wiki_url": "",
    "tracker_url": "",
    "category": "Import-Export"
}


#
# ## Declare functional classes ##
#


@bpy_extras.io_utils.orientation_helper(axis_forward='-Z', axis_up='Y')
class B2MW_OT_mw_scene_export(bpy.types.Operator, ExportHelper):
    """docstring for B2MW_OT_mw_scene_export"""
    
    bl_idname = "export_scene.where"
    bl_label = "Export MaxWhere "
    bl_options = {'PRESET'}
    
    filename_ext = ".where"
    
    # axis_forward = '-Z'
    # axis_up = 'Y'
    
    export_selected: BoolProperty(
            name="Selected Only",
            description="export only selected objects",
            default=False,
    )
    
    export_hidden: BoolProperty(
            name="Export Hidden",
            description="export hidden objects",
            default=False,
    )
    
    export_armature: BoolProperty(
            name="Armature",
            description="export bones and bone weights",
            default=False,
            )
    
    export_animation: BoolProperty(
            name="Animation",
            description="export animation frames",
            default=False,
            )
    
    export_physics: BoolProperty(
            name="Physics",
            description="export physics properties",
            default=True,
            )
    
    ex_phys_auto_props: BoolProperty(
            name="Physics auto-props",
            description="Automatic properties for physics export",
            default=True,
            )
    
    ex_phys_bias: FloatProperty(
            name="Physics property bias",
            description="",
            min=0.01, max=0.42,
            default=0.4,
            )
    
    export_materials: BoolProperty(
            name="Materials",
            description="export materials",
            default=True,
            )
    
    do_binary: BoolProperty(
            name="Binary files",
            description="create ogre-specific binary files",
            default=True,
            )
    
    do_clean: BoolProperty(
            name="Clean export folder",
            description="Delete temporary (XML) files after export",
            default=True,
    )
    
    do_unique_names: BoolProperty(
            name="Unique names",
            description="Make the auto-generated textures' names unique, for easier handling",
            default=False,
    )
    
    apply_modifiers: BoolProperty(
            name="Apply modifiers",
            description="apply modifiers from the modifier stack for each mesh",
            default=True,
            )
    
    global_scale: FloatProperty(
            name="Scale",
            description="Global scale for the scene",
            min=0.01, max=1000.0,
            default=100.0,
            )
    
    phy_mesh_quality: IntProperty(
            name="Physical mesh quality",
            description="Value between -5 and 0. -5 is the worst and 0 is the best quality.",
            min=-5, max=0,
            default=0,
    )
    
    def execute(self, context):
        from .utils_mxw import mw_export
        
        # get execution root directory (basically 'io_b2mw' as root)
        fn = inspect.getframeinfo(inspect.currentframe()).filename
        path_exec = path.dirname(path.abspath(fn))
        
        from mathutils import Matrix
        keywords = self.as_keywords(ignore=("axis_forward",
                                            "axis_up",
                                            "global_scale",
                                            "check_existing",
                                            "filter_glob",
                                            ))
        
        # global_matrix = (Matrix.Scale(self.global_scale, 4) @ axis_conversion(to_forward="-Z", to_up="Y").to_4x4())
        global_matrix = (Matrix.Scale(self.global_scale, 4)) @ axis_conversion(to_forward=self.axis_forward, to_up=self.axis_up,).to_4x4()
        
        keywords["global_matrix"] = global_matrix
        keywords["path_exec_root"] = path_exec
        keywords["sr"] = self
        
        return mw_export.save(context, **keywords)


@bpy_extras.io_utils.orientation_helper(axis_forward='-Z', axis_up='Y')
class B2MW_OT_dxt_export(bpy.types.Operator, ExportHelper):
    """ Docsstring for B2MW_OT_dxt_export """
    
    # resolution 1024
    # empty based rotation - parenting w/o inverse
    # active render properties
    # keep render png (default off)
    # use mipmaps (default on)
    
    bl_idname = "export_scene.dxt"
    bl_label = "Export DXT (.dds) maps"
    bl_options = {'PRESET'}
    
    filename_ext = ".dds"
    
    export_hidden: BoolProperty(
            name="Export Hidden",
            description="export hidden objects",
            default=False,
    )
    
    do_clean: BoolProperty(
            name="Clean render images",
            description="Delete temporary render images after export",
            default=True,
    )
    
    def execute(self, context):
        from .utils_general import dxt_export
        
        # get execution root directory (basically 'io_b2mw' as root)
        fn = inspect.getframeinfo(inspect.currentframe()).filename
        path_exec = path.dirname(path.abspath(fn))
        
        from mathutils import Matrix
        keywords = self.as_keywords(ignore=("axis_forward",
                                            "axis_up",
                                            "global_scale",
                                            "check_existing",
                                            "filter_glob",
                                            ))
        
        global_matrix = (Matrix.Scale(self.global_scale, 4) @ axis_conversion(
            to_forward=self.axis_forward, to_up=self.axis_up,).to_4x4())
        
        keywords["global_matrix"] = global_matrix
        keywords["path_exec_root"] = path_exec
        keywords["sr"] = self
        
        return dxt_export.save(context, **keywords)


@bpy_extras.io_utils.orientation_helper(axis_forward='-Z', axis_up='Y')
class B2MW_OT_physical_export(bpy.types.Operator, ExportHelper):
    """ Docsstring for B2MW_OT_physical_export """
    
    bl_idname = "export_scene.phys_ex"
    bl_label = "Export manually created physical objects"
    bl_options = {'PRESET'}
    
    filename_ext = ".phy.obj"
    
    export_selected: BoolProperty(
            name="Selected Only",
            description="export only selected objects",
            default=True,
    )
    
    do_clean: BoolProperty(
            name="Clean temporary files",
            description="Delete temporary obj files after export",
            default=True,
    )
    
    global_scale: FloatProperty(
            name="Scale",
            min=0.01, max=1000.0,
            default=1.0,
            )
    
    def execute(self, context):
        from .utils_mxw import mw_phys_ex
        
        # get execution root directory (basically 'io_b2mw' as root)
        fn = inspect.getframeinfo(inspect.currentframe()).filename
        path_exec = path.dirname(path.abspath(fn))
        
        from mathutils import Matrix
        keywords = self.as_keywords(ignore=("axis_forward",
                                            "axis_up",
                                            "global_scale",
                                            "check_existing",
                                            "filter_glob",
                                            ))
        
        global_matrix = (Matrix.Scale(self.global_scale, 4) @ axis_conversion(
            to_forward=self.axis_forward, to_up=self.axis_up,).to_4x4())
        
        keywords["global_matrix"] = global_matrix
        keywords["path_exec_root"] = path_exec
        keywords["sr"] = self
        
        return mw_phys_ex.save(context, **keywords)

#
# ## GUI STUFF + META-OPS: ##
#

# Top-level:
def menu_func_export(self, context):
    self.layout.operator(B2MW_OT_mw_scene_export.bl_idname, text="MaxWhere (.jsx, .json, .mesh, .obj, .material)")
    self.layout.operator(B2MW_OT_physical_export.bl_idname, text="MaxWhere Physical Models (_phy.obj)")
    self.layout.operator(B2MW_OT_dxt_export.bl_idname, text="DXT (.dds cube / sphere maps)")


def menu_func(self, context):
    lay_out = self.layout
    lay_out.operator_context = 'INVOKE_REGION_WIN'
    
    # Create layout (this is later used to add the VIEW3D_MT_maxwhere_add class to 'add menu' layout)
    lay_out.separator()
    lay_out.menu("VIEW3D_MT_maxwhere_add", text="MaxWhere")


# Add menu item
class VIEW3D_MT_maxwhere_add(bpy.types.Menu):
    # Define the "Extra Objects" menu
    bl_idname = "VIEW3D_MT_maxwhere_add"
    bl_label = "MaxWhere"
    
    def draw(self, context):
        layout = self.layout
        layout.operator_context = 'INVOKE_REGION_WIN'
        layout.operator(B2MW_OT_webtable_add.bl_idname, text="Webtable")
        layout.operator(B2MW_OT_dxt_add.bl_idname, text="DXT Export helper")


# Scene panel
class SCENE_PT_b2mw_panel(bpy.types.Panel):
    bl_idname = "SCENE_PT_b2mw_panel"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "scene"
    bl_label = "MaxWhere properties"
    
    def draw(self, context):
        layout = self.layout
        scene = context.scene
        
        if "mxw_props" not in scene.keys():
            layout.operator(SCENE_OT_init_mxw_props.bl_idname, text="Initialize MaxWhere Properties")
        else:
            layout.label(text="Please fill in the fields below.")
            
            layout.prop(scene, '["mxw_name"]', text="Name")
            layout.prop(scene, '["mxw_creator"]', text="Creator")
            layout.prop(scene, '["mxw_def_rfx"]', text="Default reflection map")


# Scene init op
class SCENE_OT_init_mxw_props(bpy.types.Operator):
    """ Docsstring """
    bl_idname="scene.init_mxw_props"
    bl_label="initialize mxw_props"
    
    def execute(self, context):
        
        if "mxw_props" not in bpy.context.scene.keys():
            bpy.context.scene["mxw_props"] = True
            bpy.context.scene["mxw_name"] = ""
            bpy.context.scene["mxw_creator"] = ""
            bpy.context.scene["mxw_def_rfx"] = ""
        
        return {"FINISHED"}


# Material panel
class MATERIAL_PT_b2mw_panel(bpy.types.Panel):
    bl_idname = "MATERIAL_PT_b2mw_panel"
    bl_space_type = "PROPERTIES"
    bl_region_type = "WINDOW"
    bl_context = "material"
    bl_label = "MaxWhere properties"
    
    def draw(self, context):
        layout = self.layout
        material = context.active_object.active_material
        
        if material is not None:
            if "mxw_props" not in material.keys():
                layout.operator(MATERIAL_OT_init_mxw_props.bl_idname, text="Initialize MaxWhere Properties")
            else:
                layout.label(text="Please fill in the fields below.")
                layout.prop(material, '["mxw_rfx"]', text="Reflection map")


# Material init op
class MATERIAL_OT_init_mxw_props(bpy.types.Operator):
    """ Docsstring """
    bl_idname="material.init_mxw_props"
    bl_label="initialize mxw_props"
    
    def execute(self, context):
        material = context.active_object.active_material
        
        if "mxw_props" not in material.keys():
            material["mxw_props"] = True
            material["mxw_rfx"] = ""
        
        return {"FINISHED"}


# ## ADD OPERATORS ##

# ## Webtable object ##
class B2MW_OT_webtable_add(bpy.types.Operator, AddObjectHelper):
    bl_idname = "object.b2mw_webtable_add"
    bl_label = "Webtable"
    bl_description = "Add a webtable object"
    bl_options = {'REGISTER', 'UNDO', 'PRESET'}
    
    resolutions = [
            ("-1 -1", "autoresolution", "a:r, large", "", 1),
            ("400 300", "400x300 (4:3)", "4:3 small size", "", 2),
            ("800 600", "800x600 (4:3)", "4:3 small size", "", 3),
            ("1280 720", "1280x720 (16:9)", "16:9 small size", "", 4),
            ("1600 900", "1600x900 (16:9)", "16:9 medium size", "", 5),
            ("1920 1080", "1920x1080 (16:9)", "16:9 large size", "", 6)
            ]
    
    sizes = [
            ("1.6 0.9", "160x90 cm (16:9)", "16:9 small size", "", 1),
            ("3.2 1.8", "320x180 cm (16:9)", "16:9 medium size", "", 2),
            ("4.8 2.7", "480x270 cm (16:9)", "16:9 large size", "", 3),
            ("1.2 0.9", "120x90 cm (4:3)", "4:3 small size", "", 4),
            ("2.4 1.8", "240x180 cm (4:3)", "4:3 medium size", "", 5)
            ]
    
    size: EnumProperty(
            name="Size",
            description="The physical size of the webtable",
            items=sizes
            )
    
    resolution: EnumProperty(
            name="Resolution",
            description="The resolution of the webtable",
            items=resolutions
            )
    
    make_guide: BoolProperty(
            name="Guide",
            description="Parent a helper object to the webtable, indicating its front",
            default=0
            )
    
    def execute(self, context):
        from .utils_mxw.types import mw_objs
        from mathutils import Matrix
        
        # get execution root directory (basically 'io_b2mw' as root)
        fn = inspect.getframeinfo(inspect.currentframe()).filename
        path_exec = path.dirname(path.abspath(fn))
        
        keywords = self.as_keywords(ignore=("view_align",
                                            "align",
                                            "location",
                                            "rotation"
                                            ))
        
        keywords["path_exec_root"] = path_exec
        keywords["sr"] = self
        keywords["EX_GLOBAL_MATRIX"] = (Matrix.Scale(100, 4) @ axis_conversion(to_forward="-Z", to_up="Y").to_4x4())
        
        return mw_objs.webtable_add(context, **keywords)


# ## DXT export helper object ##
class B2MW_OT_dxt_add(bpy.types.Operator, AddObjectHelper):
    bl_idname = "object.b2mw_dxt_add"
    bl_label = "DXT Cube / Spherical map"
    bl_description = "Add a DXT (.dds) cube / spherical map creation / export helper object"
    bl_options = {'REGISTER', 'UNDO', 'PRESET'}
    
    render_res: IntProperty(
            name="Render resolution",
            description="The XY size in pixels of the rendered cubemap.",
            default=1024
            )
    
    def execute(self, context):
        from .utils_general.types import dxt_objs
        
        keywords = self.as_keywords(ignore=("view_align",
                                            "align",
                                            "location",
                                            "rotation"
                                            ))
        
        keywords["sr"] = self
        
        return dxt_objs.dxt_add(context, **keywords)


#
# ## END OF GUI STUFF + META-OPS ##
#


#
# ## Listing classes, preparing for registering ##
#

classes = (
    B2MW_OT_mw_scene_export,
    B2MW_OT_dxt_export,
    B2MW_OT_physical_export,
    B2MW_OT_dxt_add,
    B2MW_OT_webtable_add,
    SCENE_PT_b2mw_panel,
    SCENE_OT_init_mxw_props,
    MATERIAL_PT_b2mw_panel,
    MATERIAL_OT_init_mxw_props,
    VIEW3D_MT_maxwhere_add
)


#
# ## Registering loop ##
#


def register():
    for cls in classes:
        bpy.utils.register_class(cls)
    
    # Add the export functions to the export menu
    bpy.types.TOPBAR_MT_file_export.append(menu_func_export)
    
    # Add "MaxWhere" menu to the "Add Object" menu and context menu.
    bpy.types.VIEW3D_MT_add.append(menu_func)


def unregister():
    # Remove the export functions to the export menu
    bpy.types.TOPBAR_MT_file_export.remove(menu_func_export)
    
    # Remove "Extras" menu from the "Add Mesh" menu and context menu.
    bpy.types.VIEW3D_MT_add.remove(menu_func)
    
    for cls in classes:
        bpy.utils.unregister_class(cls)


if __name__ == "__main__":
    register()
