import os
from os import path, listdir
import pathlib
import subprocess
import shutil
import json
import re

# from mathutils import Matrix, Vector, Color
# from bpy_extras import io_utils, node_shader_utils

import bpy
from bpy_extras.wm_utils.progress_report import (ProgressReport, ProgressReportSubstep)

from ..utils_general import zutil

#
# ## Save ##
#


def save(context,
         sr,
         filepath,
         *,
         export_hidden,
         do_clean,
         global_matrix,
         path_exec_root
         ) -> set:
    """
    The main function called from __init__.py to save DXT maps
    """
    
    import cProfile, pstats, io
    from pstats import SortKey
    pr = cProfile.Profile()
    pr.enable()
    
    with ProgressReport(context.window_manager) as progress:
        progress.enter_substeps(1)
        
        write_loop(context, sr, filepath, path_exec_root,
        EX_HIDDEN=export_hidden,
        EX_CLEAN=do_clean,
        EX_GLOBAL_MATRIX=global_matrix,
        progress=progress
        )
        
        progress.leave_substeps()
    
    pr.disable()
    s = io.StringIO()
    sortby = SortKey.CUMULATIVE
    ps = pstats.Stats(pr, stream=s).sort_stats(sortby)
    ps.print_stats(20)
    print(s.getvalue())
    
    # Return finished state; if we had errors, it would yield a traceback anyway.
    return {'FINISHED'}


#
# ## Write call ##
#


def write_loop(context, sr, filepath, path_exec_root,
               EX_HIDDEN,
               EX_CLEAN,
               EX_GLOBAL_MATRIX,
               progress: ProgressReport
               ) -> None:
    
    # WRITE LOOP BEGIN
    rtv = None
    import time
    time_start = time.time()
    
    with ProgressReportSubstep(progress, 7, "Exporting DXT maps...", "Export finished!") as subprogress1:
        
        # construct paths
        path_main, ext = path.splitext(filepath)
        
        if path.exists(path_main):
            # print(f"Warning! the path: \"{path_main}\" exists already. Deleting.")
            shutil.rmtree(path_main)
            pass
        
        paths = list()
        paths.append(path_main)
        
        for p in paths:
            os.makedirs(p)
        
        # stopper
        if not os.path.exists(path_main):
            sr.report({"ERROR"}, "Something went horribly wrong. Please make sure the export directory exists.")
            return None
        
        # ## Set rotations to be quats ##
        # cause otherwise things go wrong.
        
        for obj in bpy.data.objects:
            obj.rotation_mode = "QUATERNION"
        # just to make absolutely sure.
        context.view_layer.update()
        
        objls = []
        for obj in bpy.data.objects:
            
            pass
        
        # ## Render image(s) & export .dds ##
        
        def ex_dxt():
            
            return True
        
        rtv = ex_dxt()
        if rtv is None:
            return None
        
        # ## Clean export folder of unneded resources ##
        
        def ex_clean():
        
            def anyin(tu, st):
                for x in tu:
                    if x in st:
                        return True
                return False
            
            # init file list
            path_garbage = list()
            # fill list with files to convert
            for fp in listdir(path_dir_res):
                fp2 = path.join(path_dir_res, fp)
                gb_ext = ".mesh.xml"
                if path.isfile(fp2) and (gb_ext in fp):
                    path_garbage.append(fp2)
            
            # print(f"INFO - Garbage: {path_garbage}")
            
            for fp in path_garbage:
                zutil.rmfile(fp)
            
            return True
        
        if EX_CLEAN:
            # print(f"INFO - Cleaning export folder", flush=True)
            rtv = ex_clean()
            if rtv is None:
                return None
            subprogress1.step("Export folder cleaned.")
        else:
            subprogress1.step()
        
    
    if rtv is not None:
        sr.report({"INFO"}, f"MaxWhere export finished! Time taken: {(time.time() - time_start):.2f} s")
