from os import path

import bpy
from ...utils_general import zutil
from ...utils_general import vipsutil


class Material(object):
    """docstring"""
    
    mat_name = None
    diffuse_map = None
    normal_map = None
    roughness_map = None
    metallic_map = None
    reflection_map = None
    light_map = None
    ambientocclusion_map = None
    
    has_alpha = False
    
    gen_img_size = 8
    
    texture_offset = [1.0, 1.0]
    texture_tile = 1.0
    texture_tile_metallic = 1.0
    texture_tile_normal = 1.0
    texture_tile_roughness = 1.0
    
    refl_blur_bias = 0.3
    reflection_power = 0.5 # 1.2 for metallic
    roughness_bias = 0.0
    spec_multiplier = 0.5
    surface_reflectivity = [1.0, 1.0]
    
    
    # Constructor
    def __init__(self):
        pass
    
    def pull(self, sr, context, material, path_dir_res, path_reflection_map, EX_UNIQUE_NAMES, EX_CLEAN):
        
        self.mat_name = material.name
        # if we have a material that's using nodes (we have to have)
        if material.use_nodes:
            node_principled = None
            # get the Principled node
            # material.node_tree.active_output
            for nd in material.node_tree.nodes:
                if nd.bl_idname == "ShaderNodeBsdfPrincipled":
                    if len(nd.outputs['BSDF'].links) > 0 and (nd.outputs['BSDF'].links[0].to_node.bl_idname == "ShaderNodeOutputMaterial"):
                        node_principled = nd
            
            if node_principled is not None:
                tx_map = {}
                tx_map["Base Color"] = None
                tx_map["Normal"] = None
                tx_map["Roughness"] = None
                tx_map["Metallic"] = None
                
                clrs = []
                
                for key in tx_map:
                    # try to get the image-texture node (unsafe again)
                    node_tex = None
                    node_vec = None
                    basename = None
                    
                    if len(node_principled.inputs[key].links) > 0:
                        node_tex = node_principled.inputs[key].links[0].from_node
                        # this is so that we can jump over the 'Normal Map' node, if used.
                        if node_tex is not None:
                            if node_tex.bl_idname == "ShaderNodeNormalMap":
                                if len(node_tex.inputs[1].links) > 0:
                                    node_tex = node_tex.inputs[1].links[0].from_node
                                else:
                                    sr.report({"ERROR"}, f"No Color input found in material > {material.name} < at node > {node_tex.name} <, input > {key} <, is your Normal Map node set up correctly?")
                                    return None
                            if node_tex.bl_idname != "ShaderNodeTexImage":
                                sr.report({"ERROR"}, f"Image Texture node not found in material > {material.name} < at node > {node_principled.name} <, input > {key} <!")
                                return None
                            if node_tex.image.is_dirty:
                                sr.report({"ERROR"}, f"Texture is unsaved / modified in material > {material.name} < at node > {node_tex.name} <!")
                                return None
                            
                            # if we have a texture, copy it over
                            path_tex = node_tex.image.filepath
                            basename = zutil.cpfile(bpy.path.abspath(path_tex), path_dir_res)
                            
                            if len(node_tex.inputs[0].links) > 0:
                                node_vec = node_tex.inputs[0].links[0].from_node
                                if node_vec.vector_type != "POINT":
                                    node_vec = None
                        
                        if key == "Base Color":
                            path_img = path.join(path_dir_res, basename)
                            self.has_alpha = vipsutil.vips_is_transparent(path_img)
                            
                            if node_vec is not None:
                                self.texture_offset[0] = node_vec.inputs[1].default_value[0] # X offset
                                self.texture_offset[1] = node_vec.inputs[1].default_value[1] # Y offset
                                if zutil.isclose(node_vec.inputs[3].default_value[0], node_vec.inputs[3].default_value[1]): # checking if tiling is square
                                    self.texture_tile = node_vec.inputs[3].default_value[0] # tiling multiplier
                        elif key == "Normal":
                            if node_vec is not None:
                                if zutil.isclose(node_vec.inputs[3].default_value[0], node_vec.inputs[3].default_value[1]): # checking if tiling is square
                                    self.texture_tile_normal = node_vec.inputs[3].default_value[0] # tiling multiplier
                        elif key == "Roughness":
                            if node_vec is not None:
                                if zutil.isclose(node_vec.inputs[3].default_value[0], node_vec.inputs[3].default_value[1]): # checking if tiling is square
                                    self.texture_tile_roughness = node_vec.inputs[3].default_value[0] # tiling multiplier
                        elif key == "Metallic":
                            if node_vec is not None:
                                if zutil.isclose(node_vec.inputs[3].default_value[0], node_vec.inputs[3].default_value[1]): # checking if tiling is square
                                    self.texture_tile_metallic = node_vec.inputs[3].default_value[0] # tiling multiplier
                    
                    else:
                        # So if we aren't given a texture, we'll create them based on default values
                        input_socket = node_principled.inputs[key]
                        defval = node_principled.inputs[key].default_value
                        
                        clr = list()
                        if isinstance(input_socket, bpy.types.NodeSocketVector):
                            clr = [key, 127, 127, 255, 255]
                        elif isinstance(defval, bpy.types.bpy_prop_array):
                            r = zutil.rgbscale(defval[0])
                            g = zutil.rgbscale(defval[1])
                            b = zutil.rgbscale(defval[2])
                            clr = [key, r, g, b, 255]
                        elif isinstance(defval, float):
                            v = zutil.rgbscale(defval)
                            clr = [key, v, v, v, 255]
                        else:
                            sr.report({"ERROR"}, f"No suitable texture format to substitue value in material > {material.name} < at node > {node_principled.name} <, input > {key} <!")
                            return None
                        
                        if key == "Base Color":
                            input_socket = node_principled.inputs["Alpha"]
                            defval = node_principled.inputs["Alpha"].default_value
                            if isinstance(defval, float):
                                if defval != 1:
                                    self.has_alpha = True
                                v = zutil.rgbscale(defval)
                                clr[4] = v
                        
                        if EX_UNIQUE_NAMES:
                            if key == "Base Color":
                                clr[0] = "BaseColor"
                            basename = f"{self.mat_name}_{clr[0]}_{clr[1]:03.0f}{clr[2]:03.0f}{clr[3]:03.0f}{clr[4]:03.0f}.png"
                        else:
                            basename = f"{clr[1]:03.0f}{clr[2]:03.0f}{clr[3]:03.0f}{clr[4]:03.0f}.png"
                        
                        clrs.append(clr)
                    
                    tx_map[key] = basename
                
                self.diffuse_map = tx_map["Base Color"]
                self.normal_map = tx_map["Normal"]
                self.roughness_map = tx_map["Roughness"]
                self.metallic_map = tx_map["Metallic"]
                
                # Reflection map
                if ("mxw_rfx" in material.keys()) and (material["mxw_rfx"] != ""):
                    path_reflection_map = material["mxw_rfx"]
                    src = path.join(bpy.path.abspath("//"), path_reflection_map)
                    # if we can't find the file, we exit here.
                    if not path.exists(src):
                        sr.report({"ERROR"}, f"File > {path_reflection_map} < not found for reflection map of material > {material} <. Make sure your MxW properties data is correct.")
                        return None
                        
                    # copy the reflection map over
                    self.reflection_map = zutil.cpfile(src, path_dir_res)
                else:
                    self.reflection_map = path.basename(path_reflection_map)
                
                # Ambientocclusion & lightmap
                # set paths
                if EX_UNIQUE_NAMES:
                    self.light_map = f"{self.mat_name}_Light_255255255255.png"
                    self.ambientocclusion_map = f"{self.mat_name}_AO_255255255255.png"
                else:
                    self.light_map = "255255255255.png"
                    self.ambientocclusion_map = "255255255255.png"
                clrs.append(["Light", 255, 255, 255, 255])
                clrs.append(["AO", 255, 255, 255, 255])
                
                for clr in clrs:
                    if EX_UNIQUE_NAMES:
                        basename = f"{self.mat_name}_{clr[0]}_{clr[1]:03.0f}{clr[2]:03.0f}{clr[3]:03.0f}{clr[4]:03.0f}.png"
                    else:
                        basename = f"{clr[1]:03.0f}{clr[2]:03.0f}{clr[3]:03.0f}{clr[4]:03.0f}.png"
                    
                    path_img = path.join(path_dir_res, basename)
                    if not path.exists(path_img):
                        ln = self.gen_img_size
                        vipsutil.vips_create(path_img, ln, ln, clr[1], clr[2], clr[3], clr[4], EX_CLEAN)
            
            else:
                sr.report({"ERROR"}, f"No PrincipledBSDF node found in material > {material.name} <!")
                return None
        else:
            sr.report({"ERROR"}, f"No nodes found in material > {material.name} <!")
            return None
        
        return True
    
    #           #
    # Serialize #
    #           #
    
    def serialize(self) -> str:
        
        pad = "\t"
        alpha_strb = ""
        
        if self.has_alpha:
            alpha_strb = f"\
{pad}{pad}{pad}scene_blend alpha_blend\n\
{pad}{pad}{pad}// depth_write off\n\
{pad}{pad}{pad}// alpha_rejection greater_equal 128\n\
{pad}{pad}{pad}cull_hardware none\n\
{pad}{pad}{pad}cull_software none\n\n"
        else:
            alpha_strb = f"\
{pad}{pad}{pad}// scene_blend alpha_blend\n\
{pad}{pad}{pad}// depth_write off\n\
{pad}{pad}{pad}// alpha_rejection greater_equal 128\n\
{pad}{pad}{pad}// cull_hardware none\n\
{pad}{pad}{pad}// cull_software none\n\n"
        
        strb = \
f"// Material by B2MW\n\
\n\
material {self.mat_name}\n\
{{\n\
{pad}technique glsl\n\
{pad}{{\n\
{pad}{pad}pass mxw_pbs\n\
{pad}{pad}{{\n\
{alpha_strb}\
{pad}{pad}{pad}vertex_program_ref physicallybased_vp\n\
{pad}{pad}{pad}{{\n\
{pad}{pad}{pad}{pad}param_named texture_offset float2 {self.texture_offset[0]} {self.texture_offset[1]}\n\
{pad}{pad}{pad}{pad}param_named tile_count float {self.texture_tile}\n\
{pad}{pad}{pad}{pad}param_named tile_count_metallic float {self.texture_tile_metallic}\n\
{pad}{pad}{pad}{pad}param_named tile_count_normal float {self.texture_tile_normal}\n\
{pad}{pad}{pad}{pad}param_named tile_count_roughness float {self.texture_tile_roughness}\n\
{pad}{pad}{pad}}}\n\
{pad}{pad}{pad}\n\
{pad}{pad}{pad}fragment_program_ref physicallybased_fp\n\
{pad}{pad}{pad}{{\n\
{pad}{pad}{pad}{pad}param_named refl_blur_bias float {self.refl_blur_bias}\n\
{pad}{pad}{pad}{pad}param_named reflection_power float {self.reflection_power}\n\
{pad}{pad}{pad}{pad}param_named roughness_bias float {self.roughness_bias}\n\
{pad}{pad}{pad}{pad}param_named spec_multiplier float {self.spec_multiplier}\n\
{pad}{pad}{pad}{pad}param_named surface_reflectivity float2 {self.surface_reflectivity[0]} {self.surface_reflectivity[1]}\n\
{pad}{pad}{pad}}}\n\
{pad}{pad}{pad}\n\
{pad}{pad}{pad}\n\
{pad}{pad}{pad}texture_unit diffuse_map\n\
{pad}{pad}{pad}{{\n\
{pad}{pad}{pad}{pad}texture {self.diffuse_map}\n\
{pad}{pad}{pad}{pad}max_anisotropy 8\n\
{pad}{pad}{pad}}}\n\
{pad}{pad}{pad}\n\
{pad}{pad}{pad}texture_unit normal_map\n\
{pad}{pad}{pad}{{\n\
{pad}{pad}{pad}{pad}texture {self.normal_map}\n\
{pad}{pad}{pad}{pad}max_anisotropy 8\n\
{pad}{pad}{pad}}}\n\
{pad}{pad}{pad}texture_unit roughness_map\n\
{pad}{pad}{pad}{{\n\
{pad}{pad}{pad}{pad}texture {self.roughness_map}\n\
{pad}{pad}{pad}{pad}max_anisotropy 8\n\
{pad}{pad}{pad}}}\n\
{pad}{pad}{pad}\n\
{pad}{pad}{pad}texture_unit metallic_map\n\
{pad}{pad}{pad}{{\n\
{pad}{pad}{pad}{pad}texture {self.metallic_map}\n\
{pad}{pad}{pad}{pad}max_anisotropy 8\n\
{pad}{pad}{pad}}}\n\
{pad}{pad}{pad}\n\
{pad}{pad}{pad}texture_unit reflection_map\n\
{pad}{pad}{pad}{{\n\
{pad}{pad}{pad}{pad}texture {self.reflection_map} cubic combinedUVW\n\
{pad}{pad}{pad}{pad}max_anisotropy 8\n\
{pad}{pad}{pad}{pad}env_map cubic_reflection\n\
{pad}{pad}{pad}{pad}\n\
{pad}{pad}{pad}}}\n\
{pad}{pad}{pad}\n\
{pad}{pad}{pad}texture_unit light_map\n\
{pad}{pad}{pad}{{\n\
{pad}{pad}{pad}{pad}texture {self.light_map}\n\
{pad}{pad}{pad}{pad}max_anisotropy 8\n\
{pad}{pad}{pad}}}\n\
{pad}{pad}{pad}\n\
{pad}{pad}{pad}texture_unit ambientocclusion_map\n\
{pad}{pad}{pad}{{\n\
{pad}{pad}{pad}{pad}texture {self.ambientocclusion_map}\n\
{pad}{pad}{pad}{pad}max_anisotropy 8\n\
{pad}{pad}{pad}}}\n\
{pad}{pad}{pad}\n\
{pad}{pad}}}\n\
{pad}}}\n\
}}\n"
        
        return strb


class Skybox(object):
    """docstring"""
    
    env_offset = "1.0 1.0"
    env_scale = "1.0 1.0"
    env_tex = None
    
    # Constructor
    def __init__(self):
        pass
    
    def pull(self, sr, context, path_exec_root, path_dir_res, EX_CLEAN):
        
        # if we have a material that's using nodes (we have to have)
        if context.scene.world.use_nodes:
            world = context.scene.world
            node_env = None
            node_bg = None
            
            for nd in world.node_tree.nodes:
                if nd.bl_idname == "ShaderNodeBackground":
                    if len(nd.outputs["Background"].links) and (nd.outputs["Background"].links[0].to_node.bl_idname == "ShaderNodeOutputWorld"):
                        node_bg = nd
            
            if node_bg is not None:
                if len(node_bg.inputs["Color"].links) and (node_bg.inputs["Color"].links[0].from_node.bl_idname == "ShaderNodeTexEnvironment"):
                    node_env = node_bg.inputs["Color"].links[0].from_node
                    
                    if node_env.image:
                        img = node_env.image
                        if img.is_dirty:
                            sr.report({"ERROR"}, f"Texture is unsaved / modified in world material > {context.scene.world} < at node > {node_env.name} <!")
                            return None
                        path_tex = img.filepath
                        
                        # copy the skybox
                        src = bpy.path.abspath(path_tex)
                        basename = zutil.cpfile(src, path_dir_res)
                        
                    else:
                        sr.report({"ERROR"}, f"No texture found for world material > {context.scene.world} < at node > {node_env.name} <!")
                        return None
                else:
                    clr = list()
                    defval = node_bg.inputs["Color"].default_value
                    
                    r = int(zutil.rgbscale(defval[0]))
                    g = int(zutil.rgbscale(defval[1]))
                    b = int(zutil.rgbscale(defval[2]))
                    a = 255
                    clr = [r, g, b, a]
                    
                    basename = f"{clr[0]:03.0f}{clr[1]:03.0f}{clr[2]:03.0f}{clr[3]:03.0f}_skybox.png"
                    
                    path_img = path.join(path_dir_res, basename)
                    if not path.exists(path_img):
                        vipsutil.vips_create(path_img, 256, 128, clr[0], clr[1], clr[2], clr[3], EX_CLEAN)
                    
            else:
                sr.report({"ERROR"}, f"No background node found in world material > {context.scene.world} <!")
                return None
        else:
            sr.report({"ERROR"}, f"No nodes found in world material > {context.scene.world} <!")
            return None
        
        self.env_tex = basename
        
        return True
    
    #           #
    # Serialize #
    #           #
    
    def serialize(self) -> str:
        
        pad = "\t"
        strb = f"// Material by B2MW\n\
\n\
material FXUnlitSkyboxLatlong\n\
{{\n\
{pad}technique\n\
{pad}{{\n\
{pad}{pad}pass\n\
{pad}{pad}{{\n\
{pad}{pad}{pad}depth_check on\n\
{pad}{pad}{pad}depth_write off\n\
{pad}{pad}{pad}depth_func less_equal\n\
{pad}{pad}{pad}\n\
{pad}{pad}{pad}vertex_program_ref VSUnlitSkybox {{}}\n\
{pad}{pad}{pad}\n\
{pad}{pad}{pad}fragment_program_ref FSUnlitSkyboxLatlong\n\
{pad}{pad}{pad}{{\n\
{pad}{pad}{pad}{pad}param_named UTexCoordOffset float2 {self.env_offset}\n\
{pad}{pad}{pad}{pad}param_named UTexCoordScale float2 {self.env_scale}\n\
{pad}{pad}{pad}}}\n\
{pad}{pad}{pad}\n\
{pad}{pad}{pad}texture_unit TU0SkyboxTex\n\
{pad}{pad}{pad}{{\n\
{pad}{pad}{pad}{pad}texture {self.env_tex} numMipMaps 0\n\
{pad}{pad}{pad}{pad}tex_address_mode wrap wrap wrap\n\
{pad}{pad}{pad}{pad}filtering anisotropic\n\
{pad}{pad}{pad}{pad}max_anisotropy 8\n\
{pad}{pad}{pad}}}\n\
{pad}{pad}}}\n\
{pad}}}\n\
}}\n"
        
        return strb
